from sys import maxsize as max_int
import numpy as np
#from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
import spglib
from kpLib.lib import KPointLattice, KPointLatticeGenerator, INCLUDE_GAMMA

SYMPREC = 1e-5

def get_kpoints(
    lattice,
    fractional_coords,
    atomic_numbers,
    min_distance,
    min_total_kpoints=1,
    include_gamma="auto",
    symprec=SYMPREC,
    use_scale_factor=False,
):
    """
    Generate an optimized generalized Monkhorst-Pack K-point girds.

    This is a thin Python interface to the underlying C++ kpLib library. The heavy work is done
    by the underlying C++ library.

    Args:
        lattice (numpy.ndarray): A 3 x 3 array representing the lattice vectors. Each vector is stored
                                 in a row.
        fractional_coords (numpy.ndarray): An array of fractional atomic coordinates (as fractions
                                           of the a, b, c lattice vectors).
        atomic_numbers (numpy.ndarray): An array of atomic numbers, in the same order of the atoms'
                                        apperance in the frac_coords.
        include_gamma (str): Whether generate a gamma-centered K-point grid. Use `true` to generate
                             grids with the Gamma point. Use `false` to generate a grid without
                             the Gamma point. Use `auto` to let kpLib choose the one with the 
                             smaller number of symmetrically distinct K-points. If the 
                             Gamma-centered grid and the shifted grid
                             have the same number distinct k-point, choose the grid with the larger
                             effective minimum distance.
        symprec (float): Precision for the symmetry searching function of Spglib.
        use_scale_factor (bool): Enable the use of a scaling factor. It speeds up the generation
                                 when requiring a very dense grids (i.e. very large `min_distance`
                                 value, e.g. >100 Angstroms). When this option is enabled, kpLib
                                 will try to find the best grid for a smaller `min_distance` and
                                 then scales the grid up to satify the input requirement.
        min_distance(float): Minimum distance in real-space between any pair of lattice points in 
                             the correspondin real-space superlattice. Note this is not distance in
                             recriprocal space.
        min_total_kpoint(int): Minimum number of required total k-points.
    """
    
    # For a full list of keys in the spglib dataset, see:
    # https://spglib.readthedocs.io/en/latest/python-spglib.html#method-standardization-and-finding-primitive
    spglib_cell = (lattice, fractional_coords, atomic_numbers)
    spglib_dataset = spglib.get_symmetry_dataset(spglib_cell, symprec=symprec, angle_tolerance=-1)
    spacegroup = spglib_dataset['number']
    conventional_lattice = spglib.standardize_cell(spglib_cell, to_primitive=False, no_idealize=True,
            symprec=symprec, angle_tolerance=-1)[0]
    rotations = np.array(spglib_dataset["rotations"], dtype=np.int32)

    # For trigonal systems, we use the hP lattice setting. I.e. conventional lattice
    # is always a hexagonal lattice. The original input lattice vectors won't be
    # changed and the rotational operations are still in the basis of the input lattice vectors.
    is_conventional_hex = 143 <= spacegroup <= 194

    # The monoclinic system (2/m Laue class) should use the 2-fold axis
    # as the 3rd-vector in the conventional lattice. In international convention (used by
    # spglib), the centered lattice point in face-centered monoclinic lattice is on the a-b plane,
    # i.e. C-centered representation, while the rotational axes of the 2-fold rotation and
    # the normal of the mirror plane is along the b-vector. In our library, we always want the 
    # the direction with the highest symmetry to be along the c-vector. Therefore, the lattice
    # vectors are rotated in the order: a->b, b->c, c->a. By this rotation, the centered lattice
    # is in the new a-c plane and the new c-vector has the highest symmetry.
    if 3 <= spacegroup <= 15:
        # Note: the C-API of spglib store lattice vector as column vectors, while their python
        # interface represent them as row vectors. So rotation should be performed along the 
        # row axis.
        conventional_lattice = np.roll(conventional_lattice, 1, axis=0)
    
    # Ensure there is an inversion operator in the set
    inv_op = np.array([[-1.0, 0, 0], [0, -1.0, 0], [0, 0, -1.0]])
    if not any(np.allclose(rot, inv_op) for rot in rotations):
        inv_rotations = np.zeros((rotations.shape[0], 3, 3)).astype(int)
        for i, rot in zip(range(rotations.shape[0]), rotations):
            inv_rotations[i] = rotations[i].dot(inv_op)
        rotations = np.concatenate((rotations, inv_rotations)).astype(int)

    kpt_gen = KPointLatticeGenerator(
        spglib_cell[0],
        conventional_lattice,
        rotations.transpose(0, 2, 1),
        rotations.shape[0],
        is_conventional_hex,
    )

    include_gamma = include_gamma.lower()
    # For backward compatibility
    if include_gamma == "true" or include_gamma == "gamma" or include_gamma[0] == "g":
        kpt_gen.includeGamma(INCLUDE_GAMMA.TRUE)
    elif include_gamma == "false" or include_gamma == "shifted" or include_gamma[0] == "s":
        kpt_gen.includeGamma(INCLUDE_GAMMA.FALSE)
    else:
        kpt_gen.includeGamma(INCLUDE_GAMMA.AUTO)

    if use_scale_factor:
        kpt_gen.useScaleFactor(spacegroup)

    lattice = kpt_gen.getKPointLattice(min_distance, min_total_kpoints)

    if lattice.getNumDistinctKPoints() == max_int:
        raise Exception(
            "Error: There is a problem generating k-point grid based "
            "on your input. If you have activated scale factor, please "
            "check your request doesn't exceed the maximum allowed "
            "number of k-points."
        )

    periodic_distance = lattice.getMinPeriodicDistance()
    num_distinct_kpts = lattice.getNumDistinctKPoints()
    num_total_kpts = lattice.numTotalKPoints()
    coords = lattice.getKPointCoordinates()
    weights = lattice.getKPointWeights()
    distinct_coords, distinct_weights = zip(
        *[
            (coord, weight)
            for (coord, weight) in zip(coords, weights)
            if weight > 0
        ]
    )

    return {
        "min_periodic_distance": periodic_distance,
        "num_distinct_kpts": num_distinct_kpts,
        "num_total_kpts": num_total_kpts,
        "coords": coords,
        "weights": weights,
        "distinct_coords": distinct_coords,
        "distinct_weights": distinct_weights
    }



import click
import numpy as np
from pymatgen.core import Structure
from pymatgen.io.vasp.inputs import Kpoints, Kpoints_supported_modes
from kpLib import __version__
from kpLib.interface import get_kpoints, SYMPREC


@click.command()
@click.argument("structure", default="POSCAR", required=False)
@click.argument("output", default="KPOINTS", required=False)
@click.option(
    "-g",
    "--gamma",
    "include_gamma",
    default="auto",
    type=click.Choice(["auto", "true", "false"], case_sensitive=False),
    help="Type of Kpoint grid: auto-determined, gamma centered, or shifted grid"
)
@click.option(
    "--symprec", default=SYMPREC, help="Precision for symmetry finding in spglib."
)
@click.option(
    "-s",
    "--use_scale_factor",
    default=False,
    type=bool,
    is_flag=True,
    help="Whether use scale factor for large size grids.",
)
@click.option(
    "-d",
    "--min_distance",
    default=0.1,
    help="Minimum allowed distance of any lattice points on "
    "real-space superlattice.",
)
@click.option(
    "-n",
    "--min_total_kpoints",
    default=1,
    help="Minimum allowed number of total k-points.",
)
@click.option(
    "--version",
    "print_version",
    type=bool,
    is_flag=True,
    help="Prints out the version information for kpGen",
)
def generate(
    structure,
    output,
    min_distance,
    min_total_kpoints,
    include_gamma,
    symprec,
    use_scale_factor,
    print_version,
):
    """
    This script generates a KPOINTs output for a given structure a
    VASP POSCAR, Pymatgen structure JSON, or CIF
    """

    if print_version:
        print(f"v{__version__}")
        return

    struct = Structure.from_file(structure)

    click.echo(f"Running kpLib on {struct.composition.reduced_formula}")
    kpts = get_kpoints(
        struct.lattice.matrix,
        struct.frac_coords,
        struct.atomic_numbers,
        min_distance = min_distance,
        min_total_kpoints = min_total_kpoints,
        include_gamma = include_gamma.lower(),
        symprec = symprec,
        use_scale_factor = use_scale_factor
    )

    # Get distinct points and weights for KPOINTS file
    distinct_coords, distinct_weights = zip(
        *[
            (coord, weight)
            for (coord, weight) in zip(kpts["coords"], kpts["weights"])
            if weight > 0
        ]
    )

    click.echo(f"Generated {len(distinct_coords)} distinct kpoints")
    click.echo(f"Minimum periodic distance is {kpts['min_periodic_distance']}")

    comment = (
        f"kpLib version: {__version__}. "
        f"K-Point grid has {kpts['num_total_kpts']} total points. "
        f"Actual minimum periodic distance is {kpts['min_periodic_distance']} Angstroms. "
    )
    if any([np.allclose(d, [0.0, 0.0, 0.0]) for d in distinct_coords]):
        click.echo("Grid includes gamma point.")
        comment += " Grid includes gamma point."
    else:
        click.echo("Grid does not include gamma point.")
        comment += " Grid does not include gamma point."

    kpts_obj = Kpoints(
        comment=comment,
        style=Kpoints_supported_modes.Reciprocal,
        num_kpts=len(distinct_coords),
        kpts=distinct_coords,
        kpts_weights=distinct_weights,
    )

    kpts_obj.write_file(output)

if __name__ == "__main__":
    generate()



import glob
from pathlib import Path

from setuptools import find_packages, setup
from pybind11.setup_helpers import Pybind11Extension

ext_modules = [
    Pybind11Extension(
        "kpLib.lib", # The first arguement is the name used in at the import statement.
        ["python/kpLib/interface.cpp"] + sorted(glob.glob("./src/*.cpp")),
        include_dirs=[
            "./src/",
        ],

        #To debug, uncomment these lines to build share library with debug info.
        #extra_compile_args=['-std=c++11', '-Wall', '-Werror', '-fPIC', '-O0', '-g'],
        #extra_link_args=['-g']
    )
]

tests_requires = ["pytest", "pymatgen"]
cli_requires = ["pymatgen", "click"]

with open(Path(__file__).parent.joinpath("README.md").resolve()) as f:
    long_description = f.read()

setup(
    name="kpLib", # The name appear in PyPI. Not necessarily the one used in import statement.
    version="1.1.1",
    #use_scm_version={
    #   'version_scheme': 'post-release'
    #},
    author="Yunzhe Wang",
    author_email="ywang393@jhu.edu",
    url="https://gitlab.com/muellergroup/kplib",
    description="Library for generating highly-efficient generalized Monkhorst-Pack K-point grids" \
                " to accelerate electronic structure calculations, like DFT.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages("python"),
    package_dir={"": "python"}, # Mark "./python" as the root this package.
    ext_modules=ext_modules, # Create a external module for this package.
    install_requires = [
        "pybind11",
        "spglib",
    ],
    setup_requires = [
        "pybind11",
        #"setuptools_scm"
    ],
    tests_require = tests_requires,
    extras_require = {
        'tests' : tests_requires,
        'cli' : cli_requires,
    },
    #python_requires=">=3.7",
    #cmdclass={"build_ext": build_ext},
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "kpgen = kpLib.cli:generate [cmd]",
        ],
    },
    keywords = [
        "Computational Materials Science",
        "Materials Simulation",
        "Electronic Structure",
        "K-points",
        "kpoint",
        "Density Functional Theory",
        "DFT",
        "VASP",
        "Crystal",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="Apache-2.0",
)
